#!/usr/bin/env python3

from mpl_toolkits.basemap import Basemap
import pygraphviz as pgv
import matplotlib.pyplot as plt
import subprocess, re
import random, webcolors
import urllib.request, socket
from xml.etree import ElementTree as ETree
from collections import defaultdict

rand_color = lambda: random.choice(list(webcolors.CSS3_HEX_TO_NAMES))
pattern_ipv4 = re.compile(r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}')

def validate_ipv4(ipv4):
    '''Checks if a given string is syntactically a valid IPv4 adress
    
    validate_ipv4(str) -> bool
    
    '''
    pattern_ipv4 = re.compile(r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}')
    match_ipv4 = pattern_ipv4.fullmatch(ipv4)
    if match_ipv4:
        return True
    else:
        return False

def dns2ipv4(dns):
    '''Converts a given DNS to the string representation of the hosts IPv4 adress
    
    dns2ipv4(str) -> str

    Example: dns2ipv4("www.github.com")
    '''
    ipv4 = socket.gethostbyname(dns)
    if validate_ipv4(ipv4):
        return ipv4
    else:
        return None

def get_myinfo():
    ''' Returns geoiplookup information of the own IPv4 adress.
    
    get_hostinfo() -> dict
    
    '''
    apiurl = 'http://api.geoiplookup.net'
    hostinfo = defaultdict(str, {})
    try:
        xml = urllib.request.urlopen(apiurl).read().decode()
        xml = xml.replace('&', '')
        tree = ETree.fromstring(xml)
        for element in tree.getiterator():
            hostinfo[element.tag] = element.text
    except:
        return hostinfo
    finally:
        return hostinfo

def get_hostinfo(ipv4):
    ''' Returns geoiplookup information of agiven host adress as a dictionary.
    The adress can be given as a string representation 0f a DNS or IPv4 adress.
    
    get_hostinfo(str) -> dict
    
    Examples: get_hostinfo("www.github.com")
              get_hostinfo("151.101.12.133")
    '''

    apiurl = 'http://api.geoiplookup.net/?query='
    hostinfo = defaultdict(str, {})
    try:
        xml = urllib.request.urlopen(apiurl + dns2ipv4(ipv4)).read().decode()
        xml = xml.replace('&', '')
        tree = ETree.fromstring(xml)
        for element in tree.getiterator():
            hostinfo[element.tag] = element.text
    except:
        return hostinfo
    finally:
        return hostinfo

def ipv42geocoord(ipv4):
    '''Provisional function to map a given IPv4 adress directly to geocoordinates.

    ipv42geocoord(str) -> (float, float)

    Better use get_hostinfo directly to get all information retrived!
    '''
    hostinfo = get_hostinfo(ipv4)
    try:
        lat = hostinfo['latitude']
        lon = hostinfo['longitude']
        return (float(lat), float(lon))
    except:
        return None

def minimalmap():
    '''Draws a minimal map of the world with coastlines and countryborders

    minimalmap()

    '''
    m = Basemap()
    m.drawcoastlines()
    m.drawcountries()
    return m

def drawroutes2map(routesfile='routes.csv'):
    '''Draws all routes from the routesfile to a map of the world
    
    drawroutes2map(file)
    
    '''
    m = minimalmap()
    print('Getting Locations...')
    for route in open(routesfile).readlines():
        ips = [ipv42geocoord(ip) for ip in route.strip().split()]
        print(ips)
        locs = [loc for loc in ips if not loc == None]
        longs = [float(get_myinfo()['latitude'])] + [loc[0] for loc in locs if loc != None]
        print(longs)
        lats = [float(get_myinfo()['longitude'])] + [loc[1] for loc in locs if loc != None]
        print(lats)
        m.plot(lats, longs, '-', markersize=0, linewidth=1, color=rand_color())
    plt.savefig('world.svg', format='svg')
    plt.savefig('world.png', format='png')
    plt.show()

def drawroutes2graph(routesfile='routes.csv'):
    '''Draws all routes found in the routesfile with graphviz to a Graph
    
        drawroutes2graph(file)

    '''
    routes = open(routesfile).readlines()
    for i in range(len(routes)):
        routes[i] = routes[i].replace('*', '').split()
        G = pgv.AGraph(strict=False, directed=True)

    for l in routes:
        for i in range(len(l)-1):
            if not (l[i], l[i+1]) in set(G.edges()):
                G.add_edge(l[i], l[i+1])

    for n in G.nodes():
        if get_hostinfo(n)['countrycode'] == 'DE':
            n.attr['color'] = 'green'
        elif get_hostinfo(n)['countrycode'] == 'US':
            n.attr['color'] = 'red'
        elif get_hostinfo(n)['countrycode'] == 'ES':
            n.attr['color'] = 'yellow'
        elif get_hostinfo(n)['countrycode'] == 'CH':
            n.attr['color'] = 'blue'
        elif get_hostinfo(n)['countrycode'] == 'CN':
            n.attr['color'] = 'magenta'

        G.write('routes.dot')
        #G.layout('dot')
        #G.draw('dot.png')

        #G.layout()
        #G.draw('neato.png')

